@@ -34,6 +34,7 @@
 #include <unistd.h>
 #include <memory.h>
 #include <string.h>
+#include <pthread.h>
 
 #ifdef	malloc
 #undef	malloc
@@ -43,7 +44,7 @@
 #undef	calloc
 #endif
 
-static const char	version[] = "\n  Electric Fence 2.0.5"
+static const char	version[] = "\n  Electric Fence 2.1"
  " Copyright (C) 1987-1998 Bruce Perens.\n";
 
 /*
@@ -127,6 +128,13 @@ int		EF_PROTECT_BELOW = -1;
  */
 int		EF_ALLOW_MALLOC_0 = -1;
 
+/*
+ * EF_FREE_WIPES is set if Electric Fence is to wipe the memory content
+ * of freed blocks.  This makes it easier to check if memory is freed or
+ * not
+ */
+int            EF_FREE_WIPES = -1;
+
 /*
  * allocationList points to the array of slot structures used to manage the
  * malloc arena.
@@ -177,6 +185,35 @@ static int		noAllocationListProtection = 0;
  */
 static size_t		bytesPerPage = 0;
 
+ /*
+ * mutex to enable multithreaded operation
+ */
+static pthread_mutex_t mutex ;
+static pid_t mutexpid=0;
+static int locknr=0;
+
+
+static void lock() {
+    if (pthread_mutex_trylock(&mutex)) {
+       if (mutexpid==getpid()) {
+           locknr++;
+           return;
+       } else {
+           pthread_mutex_lock(&mutex);
+       }
+    } 
+    mutexpid=getpid();
+    locknr=1;
+}
+
+static void unlock() {
+    locknr--;
+    if (!locknr) {
+       mutexpid=0;
+       pthread_mutex_unlock(&mutex);
+    }
+}
+
 /*
  * internalError is called for those "shouldn't happen" errors in the
  * allocator.
@@ -267,6 +304,16 @@ initialize(void)
 			EF_ALLOW_MALLOC_0 = 0;
 	}
 
+	/*
+	 * See if the user wants us to wipe out freed memory.
+	 */
+	if ( EF_FREE_WIPES == -1 ) {
+	        if ( (string = getenv("EF_FREE_WIPES")) != 0 )
+	                EF_FREE_WIPES = (atoi(string) != 0);
+	        else
+	                EF_FREE_WIPES = 0;
+	}
+
 	/*
 	 * Get the run-time configuration of the virtual memory page size.
  	 */
@@ -633,8 +680,12 @@ free(void * address)
 	Slot *	previousSlot = 0;
 	Slot *	nextSlot = 0;
 
-	if ( address == 0 )
-		return;
+        lock();
+
+        if ( address == 0 ) {
+                unlock();
+                return;
+        }
 
 	if ( allocationList == 0 )
 		EF_Abort("free() called before first malloc().");
@@ -662,6 +713,9 @@ free(void * address)
 	else
 		slot->mode = FREE;
 
+       if ( EF_FREE_WIPES )
+               memset(slot->userAddress, 0xbd, slot->userSize);
+
 	previousSlot = slotForInternalAddressPreviousTo(slot->internalAddress);
 	nextSlot = slotForInternalAddress(
 	 ((char *)slot->internalAddress) + slot->internalSize);
@@ -703,6 +757,8 @@ free(void * address)
 
 	if ( !noAllocationListProtection )
 		Page_DenyAccess(allocationList, allocationListSize);
+
+        unlock();
 }
 
 extern C_LINKAGE void *
@@ -710,6 +766,8 @@ realloc(void * oldBuffer, size_t newSize)
 {
 	void *	newBuffer = malloc(newSize);
 
+        lock();
+
 	if ( oldBuffer ) {
 		size_t	size;
 		Slot *	slot;
@@ -740,6 +798,7 @@ realloc(void * oldBuffer, size_t newSize)
 		
 		/* Internal memory was re-protected in free() */
 	}
+	unlock();
 
 	return newBuffer;
 }
@@ -747,19 +806,32 @@ realloc(void * oldBuffer, size_t newSize)
 extern C_LINKAGE void *
 malloc(size_t size)
 {
-	if ( allocationList == 0 )
-		initialize();	/* This sets EF_ALIGNMENT */
+        void  *allocation;   
+ 
+        if ( allocationList == 0 ) {
+                pthread_mutex_init(&mutex, NULL); 
+                initialize();   /* This sets EF_ALIGNMENT */
+        }       
+        lock();
+        allocation=memalign(EF_ALIGNMENT, size); 
 
-	return memalign(EF_ALIGNMENT, size);
+        unlock();
+
+	return allocation;
 }
 
 extern C_LINKAGE void *
 calloc(size_t nelem, size_t elsize)
 {
 	size_t	size = nelem * elsize;
-	void *	allocation = malloc(size);
+        void * allocation;
+        
+        lock();
+       
+        allocation = malloc(size);
+        memset(allocation, 0, size);
+        unlock();
 
-	memset(allocation, 0, size);
 	return allocation;
 }
 
@@ -770,5 +842,11 @@ calloc(size_t nelem, size_t elsize)
 extern C_LINKAGE void *
 valloc (size_t size)
 {
-	return memalign(bytesPerPage, size);
+        void * allocation;
+       
+        lock();
+        allocation= memalign(bytesPerPage, size);
+        unlock();
+       
+        return allocation;
 }
